import re

from disco.bot.parser import ArgumentSet, ArgumentError
from disco.util.functional import cached_property

REGEX_FMT = '({})'
ARGS_REGEX = '( (.*)$|$)'


class CommandEvent(object):
    """
    An event which is created when a command is triggered. Contains information
    about the message, command, and parsed arguments (along with shortcuts to
    message information).

    Attributes
    ---------
    command : :class:`Command`
        The command this event was created for (aka the triggered command).
    msg : :class:`disco.types.message.Message`
        The message object which triggered this command.
    match : :class:`re.MatchObject`
        The regex match object for the command.
    name : str
        The command name (or alias) which was triggered by the command
    args : list(str)
        Arguments passed to the command
    """

    def __init__(self, command, msg, match):
        self.command = command
        self.msg = msg
        self.match = match
        self.name = self.match.group(1)
        self.args = self.match.group(2).strip().split(' ')

    @cached_property
    def member(self):
        """
        Guild member (if relevant) for the user that created the message
        """
        return self.guild.get_member(self.author)

    @property
    def channel(self):
        """
        Channel the message was created in
        """
        return self.msg.channel

    @property
    def guild(self):
        """
        Guild (if relevant) the message was created in
        """
        return self.msg.guild

    @property
    def author(self):
        """
        Author of the message
        """
        return self.msg.author


class CommandError(Exception):
    """
    An exception which is thrown when the arguments for a command are invalid,
    or don't match the commands specifications.
    """


class Command(object):
    """
    An object which defines and handles the triggering of a function based on
    user input (aka a command).

    Attributes
    ----------
    plugin : :class:`disco.bot.plugin.Plugin`
        The plugin this command is a member of.
    func : function
        The function which is called when this command is triggered.
    trigger : str
        The primary trigger (aka name).
    args : Optional[str]
        The argument format specification.
    aliases : Optional[list(str)]
        List of trigger aliases.
    group : Optional[str]
        The group this command is a member of.
    is_regex : Optional[bool]
        Whether the triggers for this command should be treated as raw regex.
    """
    def __init__(self, plugin, func, trigger, args=None, aliases=None, group=None, is_regex=False):
        self.plugin = plugin
        self.func = func
        self.triggers = [trigger] + (aliases or [])

        self.args = ArgumentSet.from_string(args or '')
        self.group = group
        self.is_regex = is_regex

    @cached_property
    def compiled_regex(self):
        """
        A compiled version of this commands regex
        """
        return re.compile(self.regex)

    @property
    def regex(self):
        """
        The regex string that defines/triggers this command
        """
        if self.is_regex:
            return REGEX_FMT.format('|'.join(self.triggers))
        else:
            group = self.group + ' ' if self.group else ''
            return REGEX_FMT.format('|'.join(['^' + group + trigger for trigger in self.triggers]) + ARGS_REGEX)

    def execute(self, event):
        """
        Handles the execution of this command given a :class:`CommandEvent`
        object.

        Returns
        -------
        bool
            Whether this command was sucessful
        """
        if len(event.args) < self.args.required_length:
            raise CommandError('{} requires {} arguments (passed {})'.format(
                event.name,
                self.args.required_length,
                len(event.args)
            ))

        try:
            args = self.args.parse(event.args)
        except ArgumentError as e:
            raise CommandError(e.message)

        return self.func(event, *args)



from disco.api.http import Routes, HTTPClient
from disco.util.logging import LoggingClass

from disco.types.user import User
from disco.types.message import Message
from disco.types.guild import Guild, GuildMember, Role
from disco.types.channel import Channel
from disco.types.invite import Invite


def optional(**kwargs):
    """
    Takes a set of keyword arguments, creating a dictionary with only the non-
    null values.

    :returns: dict
    """
    return {k: v for k, v in kwargs.items() if v is not None}


class APIClient(LoggingClass):
    """
    An abstraction over the :class:`disco.api.http.HTTPClient` that composes requests, and fits
    the models with the returned data.
    """
    def __init__(self, client):
        super(APIClient, self).__init__()

        self.client = client
        self.http = HTTPClient(self.client.config.token)

    def gateway(self, version, encoding):
        data = self.http(Routes.GATEWAY_GET).json()
        return data['url'] + '?v={}&encoding={}'.format(version, encoding)

    def channels_get(self, channel):
        r = self.http(Routes.CHANNELS_GET, dict(channel=channel))
        return Channel.create(self.client, r.json())

    def channels_modify(self, channel, **kwargs):
        r = self.http(Routes.CHANNELS_MODIFY, dict(channel=channel), json=kwargs)
        return Channel.create(self.client, r.json())

    def channels_delete(self, channel):
        r = self.http(Routes.CHANNELS_DELETE, dict(channel=channel))
        return Channel.create(self.client, r.json())

    def channels_messages_list(self, channel, around=None, before=None, after=None, limit=50):
        r = self.http(Routes.CHANNELS_MESSAGES_LIST, dict(channel=channel), params=optional(
            around=around,
            before=before,
            after=after,
            limit=limit
        ))

        return Message.create_map(self.client, r.json())

    def channels_messages_get(self, channel, message):
        r = self.http(Routes.CHANNELS_MESSAGES_GET, dict(channel=channel, message=message))
        return Message.create(self.client, r.json())

    def channels_messages_create(self, channel, content, nonce=None, tts=False):
        r = self.http(Routes.CHANNELS_MESSAGES_CREATE, dict(channel=channel), json={
            'content': content,
            'nonce': nonce,
            'tts': tts,
        })

        return Message.create(self.client, r.json())

    def channels_messages_modify(self, channel, message, content):
        r = self.http(Routes.CHANNELS_MESSAGES_MODIFY,
                dict(channel=channel, message=message),
                json={'content': content})
        return Message.create(self.client, r.json())

    def channels_messages_delete(self, channel, message):
        self.http(Routes.CHANNELS_MESSAGES_DELETE, dict(channel=channel, message=message))

    def channels_messages_delete_bulk(self, channel, messages):
        self.http(Routes.CHANNELS_MESSAGES_DELETE_BULK, dict(channel=channel), json={'messages': messages})

    def channels_permissions_modify(self, channel, permission, allow, deny, typ):
        self.http(Routes.CHANNELS_PERMISSIONS_MODIFY, dict(channel=channel, permission=permission), json={
            'allow': allow,
            'deny': deny,
            'type': typ,
        })

    def channels_permissions_delete(self, channel, permission):
        self.http(Routes.CHANNELS_PERMISSIONS_DELETE, dict(channel=channel, permission=permission))

    def channels_invites_list(self, channel):
        r = self.http(Routes.CHANNELS_INVITES_LIST, dict(channel=channel))
        return Invite.create_map(self.client, r.json())

    def channels_invites_create(self, channel, max_age=86400, max_uses=0, temporary=False, unique=False):
        r = self.http(Routes.CHANNELS_INVITES_CREATE, dict(channel=channel), json={
            'max_age': max_age,
            'max_uses': max_uses,
            'temporary': temporary,
            'unique': unique
        })
        return Invite.create(self.client, r.json())

    def channels_pins_list(self, channel):
        r = self.http(Routes.CHANNELS_PINS_LIST, dict(channel=channel))
        return Message.create_map(self.client, r.json())

    def channels_pins_create(self, channel, message):
        self.http(Routes.CHANNELS_PINS_CREATE, dict(channel=channel, message=message))

    def channels_pins_delete(self, channel, message):
        self.http(Routes.CHANNELS_PINS_DELETE, dict(channel=channel, message=message))

    def guilds_get(self, guild):
        r = self.http(Routes.GUILDS_GET, dict(guild=guild))
        return Guild.create(self.client, r.json())

    def guilds_modify(self, guild, **kwargs):
        r = self.http(Routes.GUILDS_MODIFY, dict(guild=guild), json=kwargs)
        return Guild.create(self.client, r.json())

    def guilds_delete(self, guild):
        r = self.http(Routes.GUILDS_DELETE, dict(guild=guild))
        return Guild.create(self.client, r.json())

    def guilds_channels_list(self, guild):
        r = self.http(Routes.GUILDS_CHANNELS_LIST, dict(guild=guild))
        return Channel.create_map(self.client, r.json())

    def guilds_channels_create(self, guild, **kwargs):
        r = self.http(Routes.GUILDS_CHANNELS_CREATE, dict(guild=guild), json=kwargs)
        return Channel.create(self.client, r.json())

    def guilds_channels_modify(self, guild, channel, position):
        self.http(Routes.GUILDS_CHANNELS_MODIFY, dict(guild=guild), json={
            'id': channel,
            'position': position,
        })

    def guilds_members_list(self, guild):
        r = self.http(Routes.GUILDS_MEMBERS_LIST, dict(guild=guild))
        return GuildMember.create_map(self.client, r.json())

    def guilds_members_get(self, guild, member):
        r = self.http(Routes.GUILD_MEMBERS_GET, dict(guild=guild, member=member))
        return GuildMember.create(self.client, r.json())

    def guilds_members_modify(self, guild, member, **kwargs):
        self.http(Routes.GUILD_MEMBERS_MODIFY, dict(guild=guild, member=member), json=kwargs)

    def guilds_members_kick(self, guild, member):
        self.http(Routes.GUILD_MEMBERS_KICK, dict(guild=guild, member=member))

    def guilds_bans_list(self, guild):
        r = self.http(Routes.GUILD_BANS_LIST, dict(guild=guild))
        return User.create_map(self.client, r.json())

    def guilds_bans_create(self, guild, user, delete_message_days):
        self.http(Routes.GUILDS_BANS_CREATE, dict(guild=guild, user=user), params={
            'delete-message-days': delete_message_days,
        })

    def guilds_bans_delete(self, guild, user):
        self.http(Routes.GUILDS_BANS_DELETE, dict(guild=guild, user=user))

    def guilds_roles_list(self, guild):
        r = self.http(Routes.GUILDS_ROLES_LIST, dict(guild=guild))
        return Role.create_map(self.client, r.json())

    def guilds_roles_create(self, guild):
        r = self.http(Routes.GUILDS_ROLES_CREATE, dict(guild=guild))
        return Role.create(self.client, r.json())

    def guilds_roles_modify_batch(self, guild, roles):
        r = self.http(Routes.GUILDS_ROLES_MODIFY_BATCH, dict(guild=guild), json=roles)
        return Role.create_map(self.client, r.json())

    def guilds_roles_modify(self, guild, role, **kwargs):
        r = self.http(Routes.GUILDS_ROLES_MODIFY, dict(guild=guild, role=role), json=kwargs)
        return Role.create(self.client, r.json())

    def guilds_roles_delete(self, guild, role):
        self.http(Routes.GUILDS_ROLES_DELETE, dict(guild=guild, role=role))



import inspect
import functools

from disco.util.logging import LoggingClass
from disco.bot.command import Command, CommandError


class PluginDeco(object):
    """
    A utility mixin which provides various function decorators that a plugin
    author can use to create bound event/command handlers.
    """
    @staticmethod
    def add_meta_deco(meta):
        def deco(f):
            if not hasattr(f, 'meta'):
                f.meta = []

            f.meta.append(meta)

            return f
        return deco

    @classmethod
    def listen(cls, event_name):
        """
        Binds the function to listen for a given event name
        """
        return cls.add_meta_deco({
            'type': 'listener',
            'event_name': event_name,
        })

    @classmethod
    def command(cls, *args, **kwargs):
        """
        Creates a new command attached to the function
        """
        return cls.add_meta_deco({
            'type': 'command',
            'args': args,
            'kwargs': kwargs,
        })

    @classmethod
    def pre_command(cls):
        """
        Runs a function before a command is triggered
        """
        return cls.add_meta_deco({
            'type': 'pre_command',
        })

    @classmethod
    def post_command(cls):
        """
        Runs a function after a command is triggered
        """
        return cls.add_meta_deco({
            'type': 'post_command',
        })

    @classmethod
    def pre_listener(cls):
        """
        Runs a function before a listener is triggered
        """
        return cls.add_meta_deco({
            'type': 'pre_listener',
        })

    @classmethod
    def post_listener(cls):
        """
        Runs a function after a listener is triggered
        """
        return cls.add_meta_deco({
            'type': 'post_listener',
        })


class Plugin(LoggingClass, PluginDeco):
    """
    A plugin is a set of listeners/commands which can be loaded/unloaded by a bot.

    Parameters
    ----------
    bot : :class:`disco.bot.Bot`
        The bot this plugin is a member of.
    config : any
        The configuration data for this plugin.

    Attributes
    ----------
    client : :class:`disco.client.Client`
        An alias to the client the bot is running with.
    state : :class:`disco.state.State`
        An alias to the state object for the client.
    listeners : list
        List of all bound listeners this plugin owns.
    commands : list(:class:`disco.bot.command.Command`)
        List of all commands this plugin owns.
    """
    def __init__(self, bot, config):
        super(Plugin, self).__init__()
        self.bot = bot
        self.client = bot.client
        self.state = bot.client.state
        self.config = config

        self.listeners = []
        self.commands = {}

        self._pre = {'command': [], 'listener': []}
        self._post = {'command': [], 'listener': []}

        for name, member in inspect.getmembers(self, predicate=inspect.ismethod):
            if hasattr(member, 'meta'):
                for meta in member.meta:
                    if meta['type'] == 'listener':
                        self.register_listener(member, meta['event_name'])
                    elif meta['type'] == 'command':
                        self.register_command(member, *meta['args'], **meta['kwargs'])
                    elif meta['type'].startswith('pre_') or meta['type'].startswith('post_'):
                        when, typ = meta['type'].split('_', 1)
                        self.register_trigger(typ, when, member)

    def execute(self, event):
        """
        Executes a CommandEvent this plugin owns
        """
        try:
            return event.command.execute(event)
        except CommandError as e:
            event.msg.reply(e.message)
            return False

    def register_trigger(self, typ, when, func):
        """
        Registers a trigger
        """
        getattr(self, '_' + when)[typ].append(func)

    def _dispatch(self, typ, func, event, *args, **kwargs):
        for pre in self._pre[typ]:
            event = pre(event, args, kwargs)

        if event is None:
            return False

        result = func(event, *args, **kwargs)

        for post in self._post[typ]:
            post(event, args, kwargs, result)

        return True

    def register_listener(self, func, name):
        """
        Registers a listener

        Parameters
        ----------
        func : function
            The function to be registered.
        name : string
            Name of event to listen for.
        """
        func = functools.partial(self._dispatch, 'listener', func)
        self.listeners.append(self.bot.client.events.on(name, func))

    def register_command(self, func, *args, **kwargs):
        """
        Registers a command

        Parameters
        ----------
        func : function
            The function to be registered.
        args
            Arguments to pass onto the :class:`disco.bot.command.Command` object.
        kwargs
            Keyword arguments to pass onto the :class:`disco.bot.command.Command`
            object.
        """
        wrapped = functools.partial(self._dispatch, 'command', func)
        self.commands[func.__name__] = Command(self, wrapped, *args, **kwargs)

    def destroy(self):
        """
        Destroys the plugin (removing all listeners)
        """
        map(lambda k: k.remove(), self._events)

    def load(self):
        """
        Called when the plugin is loaded
        """
        pass

    def unload(self):
        """
        Called when the plugin is unloaded
        """
        pass



from disco.types.base import Model, snowflake


class VoiceState(Model):
    session_id = str
    guild_id = snowflake
    channel_id = snowflake
    user_id = snowflake
    deaf = bool
    mute = bool
    self_deaf = bool
    self_mute = bool
    suppress = bool

    @property
    def guild(self):
        return self.client.state.guilds.get(self.guild_id)

    @property
    def channel(self):
        return self.client.state.channels.get(self.channel_id)

    @property
    def user(self):
        return self.client.state.users.get(self.user_id)



import inflection
import six

from disco.types import Guild, Channel, User, GuildMember, Role, Message, VoiceState
from disco.types.base import Model, snowflake, alias, listof


# TODO: clean this... use BaseType, etc
class GatewayEvent(Model):
    @staticmethod
    def from_dispatch(client, data):
        cls = globals().get(inflection.camelize(data['t'].lower()))
        if not cls:
            raise Exception('Could not find cls for {}'.format(data['t']))

        return cls.create(data['d'], client)

    @classmethod
    def create(cls, obj, client):
        # If this event is wrapping a model, pull its fields
        if hasattr(cls, '_wraps_model'):
            alias, model = cls._wraps_model

            data = {
                k: obj.pop(k) for k in six.iterkeys(model._fields) if k in obj
            }

            obj[alias] = data

        return cls(obj, client)


def wraps_model(model, alias=None):
    alias = alias or model.__name__.lower()

    def deco(cls):
        cls._fields[alias] = model
        cls._wraps_model = (alias, model)
        return cls
    return deco


class Ready(GatewayEvent):
    version = alias(int, 'v')
    session_id = str
    user = User
    guilds = listof(Guild)


class Resumed(GatewayEvent):
    pass


@wraps_model(Guild)
class GuildCreate(GatewayEvent):
    unavailable = bool


@wraps_model(Guild)
class GuildUpdate(GatewayEvent):
    pass


class GuildDelete(GatewayEvent):
    id = snowflake
    unavailable = bool


@wraps_model(Channel)
class ChannelCreate(GatewayEvent):
    @property
    def guild(self):
        return self.channel.guild


@wraps_model(Channel)
class ChannelUpdate(ChannelCreate):
    pass


@wraps_model(Channel)
class ChannelDelete(ChannelCreate):
    pass


class ChannelPinsUpdate(GatewayEvent):
    channel_id = snowflake
    last_pin_timestamp = int


@wraps_model(User)
class GuildBanAdd(GatewayEvent):
    pass


@wraps_model(User)
class GuildBanRemove(GuildBanAdd):
    pass


class GuildEmojisUpdate(GatewayEvent):
    pass


class GuildIntegrationsUpdate(GatewayEvent):
    pass


class GuildMembersChunk(GatewayEvent):
    guild_id = snowflake
    members = listof(GuildMember)


@wraps_model(GuildMember, alias='member')
class GuildMemberAdd(GatewayEvent):
    pass


class GuildMemberRemove(GatewayEvent):
    guild_id = snowflake
    user = User


class GuildMemberUpdate(GatewayEvent):
    guild_id = snowflake
    user = User
    roles = listof(snowflake)


class GuildRoleCreate(GatewayEvent):
    guild_id = snowflake
    role = Role


class GuildRoleUpdate(GuildRoleCreate):
    pass


class GuildRoleDelete(GuildRoleCreate):
    pass


@wraps_model(Message)
class MessageCreate(GatewayEvent):
    @property
    def channel(self):
        return self.message.channel


@wraps_model(Message)
class MessageUpdate(MessageCreate):
    pass


class MessageDelete(GatewayEvent):
    id = snowflake
    channel_id = snowflake


class MessageDeleteBulk(GatewayEvent):
    channel_id = snowflake
    ids = listof(snowflake)


class PresenceUpdate(GatewayEvent):
    class Game(Model):
        # TODO enum
        type = int
        name = str
        url = str

    user = User
    guild_id = snowflake
    roles = listof(snowflake)
    game = Game
    status = str


class TypingStart(GatewayEvent):
    channel_id = snowflake
    user_id = snowflake
    timestamp = snowflake


@wraps_model(VoiceState, alias='state')
class VoiceStateUpdate(GatewayEvent):
    pass


class VoiceServerUpdate(GatewayEvent):
    token = str
    endpoint = str
    guild_id = snowflake



from disco.types.channel import Channel
from disco.types.guild import Guild, GuildMember, Role
from disco.types.user import User
from disco.types.message import Message
from disco.types.voice import VoiceState



import requests
import random
import gevent

from holster.enum import Enum

from disco.util.logging import LoggingClass
from disco.api.ratelimit import RateLimiter

# Enum of all HTTP methods used
HTTPMethod = Enum(
    GET='GET',
    POST='POST',
    PUT='PUT',
    PATCH='PATCH',
    DELETE='DELETE',
)


class Routes(object):
    """
    Simple Python object-enum of all method/url route combinations available to
    this client.
    """
    # Gateway
    GATEWAY_GET = (HTTPMethod.GET, '/gateway')

    # Channels
    CHANNELS = '/channels/{channel}'
    CHANNELS_GET = (HTTPMethod.GET, CHANNELS)
    CHANNELS_MODIFY = (HTTPMethod.PATCH, CHANNELS)
    CHANNELS_DELETE = (HTTPMethod.DELETE, CHANNELS)

    CHANNELS_MESSAGES_LIST = (HTTPMethod.GET, CHANNELS + '/messages')
    CHANNELS_MESSAGES_GET = (HTTPMethod.GET, CHANNELS + '/messages/{message}')
    CHANNELS_MESSAGES_CREATE = (HTTPMethod.POST, CHANNELS + '/messages')
    CHANNELS_MESSAGES_MODIFY = (HTTPMethod.PATCH, CHANNELS + '/messages/{message}')
    CHANNELS_MESSAGES_DELETE = (HTTPMethod.DELETE, CHANNELS + '/messages/{message}')
    CHANNELS_MESSAGES_DELETE_BULK = (HTTPMethod.POST, CHANNELS + '/messages/bulk_delete')

    CHANNELS_PERMISSIONS_MODIFY = (HTTPMethod.PUT, CHANNELS + '/permissions/{permission}')
    CHANNELS_PERMISSIONS_DELETE = (HTTPMethod.DELETE, CHANNELS + '/permissions/{permission}')
    CHANNELS_INVITES_LIST = (HTTPMethod.GET, CHANNELS + '/invites')
    CHANNELS_INVITES_CREATE = (HTTPMethod.POST, CHANNELS + '/invites')

    CHANNELS_PINS_LIST = (HTTPMethod.GET, CHANNELS + '/pins')
    CHANNELS_PINS_CREATE = (HTTPMethod.PUT, CHANNELS + '/pins/{pin}')
    CHANNELS_PINS_DELETE = (HTTPMethod.DELETE, CHANNELS + '/pins/{pin}')

    # Guilds
    GUILDS = '/guilds/{guild}'
    GUILDS_GET = (HTTPMethod.GET, GUILDS)
    GUILDS_MODIFY = (HTTPMethod.PATCH, GUILDS)
    GUILDS_DELETE = (HTTPMethod.DELETE, GUILDS)
    GUILDS_CHANNELS_LIST = (HTTPMethod.GET, GUILDS + '/channels')
    GUILDS_CHANNELS_CREATE = (HTTPMethod.POST, GUILDS + '/channels')
    GUILDS_CHANNELS_MODIFY = (HTTPMethod.PATCH, GUILDS + '/channels')
    GUILDS_MEMBERS_LIST = (HTTPMethod.GET, GUILDS + '/members')
    GUILDS_MEMBERS_GET = (HTTPMethod.GET, GUILDS + '/members/{member}')
    GUILDS_MEMBERS_MODIFY = (HTTPMethod.PATCH, GUILDS + '/members/{member}')
    GUILDS_MEMBERS_KICK = (HTTPMethod.DELETE, GUILDS + '/members/{member}')
    GUILDS_BANS_LIST = (HTTPMethod.GET, GUILDS + '/bans')
    GUILDS_BANS_CREATE = (HTTPMethod.PUT, GUILDS + '/bans/{user}')
    GUILDS_BANS_DELETE = (HTTPMethod.DELETE, GUILDS + '/bans/{user}')
    GUILDS_ROLES_LIST = (HTTPMethod.GET, GUILDS + '/roles')
    GUILDS_ROLES_CREATE = (HTTPMethod.GET, GUILDS + '/roles')
    GUILDS_ROLES_MODIFY_BATCH = (HTTPMethod.PATCH, GUILDS + '/roles')
    GUILDS_ROLES_MODIFY = (HTTPMethod.PATCH, GUILDS + '/roles/{role}')
    GUILDS_ROLES_DELETE = (HTTPMethod.DELETE, GUILDS + '/roles/{role}')
    GUILDS_PRUNE_COUNT = (HTTPMethod.GET, GUILDS + '/prune')
    GUILDS_PRUNE_BEGIN = (HTTPMethod.POST, GUILDS + '/prune')
    GUILDS_VOICE_REGIONS_LIST = (HTTPMethod.GET, GUILDS + '/regions')
    GUILDS_INVITES_LIST = (HTTPMethod.GET, GUILDS + '/invites')
    GUILDS_INTEGRATIONS_LIST = (HTTPMethod.GET, GUILDS + '/integrations')
    GUILDS_INTEGRATIONS_CREATE = (HTTPMethod.POST, GUILDS + '/integrations')
    GUILDS_INTEGRATIONS_MODIFY = (HTTPMethod.PATCH, GUILDS + '/integrations/{integration}')
    GUILDS_INTEGRATIONS_DELETE = (HTTPMethod.DELETE, GUILDS + '/integrations/{integration}')
    GUILDS_INTEGRATIONS_SYNC = (HTTPMethod.POST, GUILDS + '/integrations/{integration}/sync')
    GUILDS_EMBED_GET = (HTTPMethod.GET, GUILDS + '/embed')
    GUILDS_EMBED_MODIFY = (HTTPMethod.PATCH, GUILDS + '/embed')

    # Users
    USERS = '/users'
    USERS_ME_GET = (HTTPMethod.GET, USERS + '/@me')
    USERS_ME_PATCH = (HTTPMethod.PATCH, USERS + '/@me')
    USERS_ME_GUILDS_LIST = (HTTPMethod.GET, USERS + '/@me/guilds')
    USERS_ME_GUILDS_LEAVE = (HTTPMethod.DELETE, USERS + '/@me/guilds/{guild}')
    USERS_ME_DMS_LIST = (HTTPMethod.GET, USERS + '/@me/channels')
    USERS_ME_DMS_CREATE = (HTTPMethod.POST, USERS + '/@me/channels')
    USERS_ME_CONNECTIONS_LIST = (HTTPMethod.GET, USERS + '/@me/connections')
    USERS_GET = (HTTPMethod.GET, USERS + '/{user}')


class APIException(Exception):
    """
    Exception thrown when an HTTP-client level error occurs. Usually this will
    be a non-success status-code, or a transient network issue.
    """
    def __init__(self, msg, status_code=0, content=None):
        self.status_code = status_code
        self.content = content
        self.msg = msg

        if self.status_code:
            self.msg += ' code: {}'.format(status_code)

        super(APIException, self).__init__(self.msg)


class HTTPClient(LoggingClass):
    """
    A simple HTTP client which wraps the requests library, adding support for
    Discords rate-limit headers, authorization, and request/response validation.
    """
    BASE_URL = 'https://discordapp.com/api/v6'
    MAX_RETRIES = 5

    def __init__(self, token):
        super(HTTPClient, self).__init__()

        self.limiter = RateLimiter()
        self.headers = {
            'Authorization': 'Bot ' + token,
        }

    def __call__(self, route, args=None, **kwargs):
        return self.call(route, args, **kwargs)

    def call(self, route, args=None, **kwargs):
        """
        Makes a request to the given route (as specified in
        :class:`disco.api.http.Routes`) with a set of URL arguments, and keyword
        arguments passed to requests.

        Parameters
        ----------
        route : tuple(:class:`HTTPMethod`, str)
            The method.URL combination that when compiled with URL arguments
            creates a requestable route which the HTTPClient will make the
            request too.
        args : dict(str, str)
            A dictionary of URL arguments that will be compiled with the raw URL
            to create the requestable route. The HTTPClient uses this to track
            rate limits as well.
        kwargs : dict
            Keyword arguments that will be passed along to the requests library

        Raises
        ------
        APIException
            Raised when an unrecoverable error occurs, or when we've exhausted
            the number of retries.

        Returns
        -------
        :class:`requests.Response`
            The response object for the request
        """
        args = args or {}
        retry = kwargs.pop('retry_number', 0)

        # Merge or set headers
        if 'headers' in kwargs:
            kwargs['headers'].update(self.headers)
        else:
            kwargs['headers'] = self.headers

        # Build the bucket URL
        filtered = {k: (v if v in ('guild', 'channel') else '') for k, v in args.items()}
        bucket = (route[0].value, route[1].format(**filtered))

        # Possibly wait if we're rate limited
        self.limiter.check(bucket)

        # Make the actual request
        url = self.BASE_URL + route[1].format(**args)
        r = requests.request(route[0].value, url, **kwargs)

        # Update rate limiter
        self.limiter.update(bucket, r)

        # If we got a success status code, just return the data
        if r.status_code < 400:
            return r
        elif r.status_code != 429 and 400 < r.status_code < 500:
            raise APIException('Request failed', r.status_code, r.content)
        else:
            if r.status_code == 429:
                self.log.warning('Request responded w/ 429, retrying (but this should not happen, check your clock sync')

            # If we hit the max retries, throw an error
            retry += 1
            if retry > self.MAX_RETRIES:
                self.log.error('Failing request, hit max retries')
                raise APIException('Request failed after {} attempts'.format(self.MAX_RETRIES), r.status_code, r.content)

            backoff = self.random_backoff()
            self.log.warning('Request to `{}` failed with code {}, retrying after {}s'.format(url, r.status_code, backoff))
            gevent.sleep(backoff)

            # Otherwise just recurse and try again
            return self(route, args, retry_number=retry, **kwargs)

    @staticmethod
    def random_backoff():
        """
        Returns a random backoff (in milliseconds) to be used for any error the
        client suspects is transient. Will always return a value between 500 and
        5000 milliseconds.

        :returns: a random backoff in milliseconds
        :rtype: float
        """
        return random.randint(500, 5000) / 1000.0



import gevent
import zlib
import six

from disco.gateway.packets import OPCode
from disco.gateway.events import GatewayEvent
from disco.gateway.encoding.json import JSONEncoder
from disco.util.websocket import Websocket
from disco.util.logging import LoggingClass

TEN_MEGABYTES = 10490000


class GatewayClient(LoggingClass):
    GATEWAY_VERSION = 6
    MAX_RECONNECTS = 5

    def __init__(self, client, encoder=None):
        super(GatewayClient, self).__init__()
        self.client = client
        self.encoder = encoder or JSONEncoder

        self.events = client.events
        self.packets = client.packets

        # Create emitter and bind to gateway payloads
        self.packets.on(OPCode.DISPATCH, self.handle_dispatch)
        self.packets.on(OPCode.HEARTBEAT, self.handle_heartbeat)
        self.packets.on(OPCode.RECONNECT, self.handle_reconnect)
        self.packets.on(OPCode.INVALID_SESSION, self.handle_invalid_session)
        self.packets.on(OPCode.HELLO, self.handle_hello)

        # Bind to ready payload
        self.events.on('Ready', self.on_ready)

        # Websocket connection
        self.ws = None

        # State
        self.seq = 0
        self.session_id = None
        self.reconnects = 0
        self.shutting_down = False

        # Cached gateway URL
        self._cached_gateway_url = None

        # Heartbeat
        self._heartbeat_task = None

    def send(self, op, data):
        self.ws.send(self.encoder.encode({
            'op': op.value,
            'd': data,
        }), self.encoder.OPCODE)

    def heartbeat_task(self, interval):
        while True:
            self.send(OPCode.HEARTBEAT, self.seq)
            gevent.sleep(interval / 1000)

    def handle_dispatch(self, packet):
        obj = GatewayEvent.from_dispatch(self.client, packet)
        self.log.debug('Dispatching %s', obj.__class__.__name__)
        self.client.events.emit(obj.__class__.__name__, obj)

    def handle_heartbeat(self, packet):
        self.send(OPCode.HEARTBEAT, self.seq)

    def handle_reconnect(self, packet):
        self.log.warning('Received RECONNECT request, forcing a fresh reconnect')
        self.session_id = None
        self.ws.close()

    def handle_invalid_session(self, packet):
        self.log.warning('Recieved INVALID_SESSION, forcing a fresh reconnect')
        self.session_id = None
        self.ws.close()

    def handle_hello(self, packet):
        self.log.info('Recieved HELLO, starting heartbeater...')
        self._heartbeat_task = gevent.spawn(self.heartbeat_task, packet['d']['heartbeat_interval'])

    def on_ready(self, ready):
        self.log.info('Recieved READY')
        self.session_id = ready.session_id
        self.reconnects = 0

    def connect_and_run(self):
        if not self._cached_gateway_url:
            self._cached_gateway_url = self.client.api.gateway(
                version=self.GATEWAY_VERSION,
                encoding=self.encoder.TYPE)

        self.log.info('Opening websocket connection to URL `%s`', self._cached_gateway_url)
        self.ws = Websocket(self._cached_gateway_url)
        self.ws.emitter.on('on_open', self.on_open)
        self.ws.emitter.on('on_error', self.on_error)
        self.ws.emitter.on('on_close', self.on_close)
        self.ws.emitter.on('on_message', self.on_message)

        self.ws.run_forever()

    def on_message(self, msg):
        # Detect zlib and decompress
        is_erlpack = ((six.PY2 and ord(msg[0]) == 131) or (six.PY3 and msg[0] == 131))
        if msg[0] != '{' and not is_erlpack:
            msg = zlib.decompress(msg, 15, TEN_MEGABYTES).decode("utf-8")

        try:
            data = self.encoder.decode(msg)
        except:
            self.log.exception('Failed to parse gateway message: ')
            return

        # Update sequence
        if data['s'] and data['s'] > self.seq:
            self.seq = data['s']

        # Emit packet
        self.packets.emit(OPCode[data['op']], data)

    def on_error(self, error):
        if isinstance(error, KeyboardInterrupt):
            self.shutting_down = True
        raise Exception('WS recieved error: %s', error)

    def on_open(self):
        if self.seq and self.session_id:
            self.log.info('WS Opened: attempting resume w/ SID: %s SEQ: %s', self.session_id, self.seq)
            self.send(OPCode.RESUME, {
                'token': self.client.config.token,
                'session_id': self.session_id,
                'seq': self.seq
            })
        else:
            self.log.info('WS Opened: sending identify payload')
            self.send(OPCode.IDENTIFY, {
                'token': self.client.config.token,
                'compress': True,
                'large_threshold': 250,
                'shard': [
                    self.client.config.shard_id,
                    self.client.config.shard_count,
                ],
                'properties': {
                    '$os': 'linux',
                    '$browser': 'disco',
                    '$device': 'disco',
                    '$referrer': '',
                }
            })

    def on_close(self, code, reason):
        if self.shutting_down:
            self.log.info('WS Closed: shutting down')
            return

        self.reconnects += 1
        self.log.info('WS Closed: [%s] %s (%s)', code, reason, self.reconnects)

        if self.MAX_RECONNECTS and self.reconnects > self.MAX_RECONNECTS:
            raise Exception('Failed to reconect after {} attempts, giving up'.format(self.MAX_RECONNECTS))

        # Don't resume for these error codes
        if code and 4000 <= code <= 4010:
            self.session_id = None

        wait_time = self.reconnects * 5
        self.log.info('Will attempt to %s after %s seconds', 'resume' if self.session_id else 'reconnect', wait_time)
        gevent.sleep(wait_time)

        # Reconnect
        self.connect_and_run()

    def run(self):
        self.connect_and_run()



from holster.enum import Enum

from disco.types.base import Model, snowflake, enum, listof, dictof, alias, text
from disco.types.permissions import PermissionValue

from disco.util.functional import cached_property
from disco.types.user import User
from disco.types.permissions import Permissions, Permissible
from disco.voice.client import VoiceClient


ChannelType = Enum(
    GUILD_TEXT=0,
    DM=1,
    GUILD_VOICE=2,
    GROUP_DM=3,
)

PermissionOverwriteType = Enum(
    ROLE='role',
    MEMBER='member'
)


class PermissionOverwrite(Model):
    """
    A PermissionOverwrite for a :class:`Channel`


    Attributes
    ----------
    id : snowflake
        The overwrite ID
    type : :const:`disco.types.channel.PermissionsOverwriteType`
        The overwrite type
    allowed : :class:`PermissionValue`
        All allowed permissions
    denied : :class:`PermissionValue`
        All denied permissions
    """

    id = snowflake
    type = enum(PermissionOverwriteType)
    allow = PermissionValue
    deny = PermissionValue


class Channel(Model, Permissible):
    """
    Represents a Discord Channel

    Attributes
    ----------
    id : snowflake
        The channel ID.
    guild_id : Optional[snowflake]
        The guild id this channel is part of.
    name : str
        The channels name.
    topic : str
        The channels topic.
    position : int
        The channels position.
    bitrate : int
        The channels bitrate.
    recipients: list(:class:`disco.types.user.User`)
        Members of this channel (if this is a DM channel).
    type : :const:`ChannelType`
        The type of this channel.
    overwrites : dict(snowflake, :class:`disco.types.channel.PermissionOverwrite`)
        Channel permissions overwrites.
    """
    id = snowflake
    guild_id = snowflake
    name = text
    topic = text
    _last_message_id = alias(snowflake, 'last_message_id')
    position = int
    bitrate = int
    recipients = listof(User)
    type = enum(ChannelType)
    overwrites = alias(dictof(PermissionOverwrite, key='id'), 'permission_overwrites')

    def get_permissions(self, user):
        """
        Get the permissions a user has in the channel

        Returns
        -------
        :class:`disco.types.permissions.PermissionValue`
            Computed permission value for the user.
        """
        if not self.guild_id:
            return Permissions.ADMINISTRATOR

        member = self.guild.members.get(user.id)
        base = self.guild.get_permissions(user)

        for ow in self.overwrites.values():
            if ow.id != user.id and ow.id not in member.roles:
                continue

            base -= ow.deny
            base += ow.allow

        return base

    @property
    def is_guild(self):
        """
        Whether this channel belongs to a guild
        """
        return self.type in (ChannelType.GUILD_TEXT, ChannelType.GUILD_VOICE)

    @property
    def is_dm(self):
        """
        Whether this channel is a DM (does not belong to a guild)
        """
        return self.type in (ChannelType.DM, ChannelType.GROUP_DM)

    @property
    def is_voice(self):
        """
        Whether this channel supports voice
        """
        return self.type in (ChannelType.GUILD_VOICE, ChannelType.GROUP_DM)

    @property
    def last_message_id(self):
        """
        Returns the ID of the last message sent in this channel
        """
        if self.id not in self.client.state.messages:
            return self._last_message_id
        return self.client.state.messages[self.id][-1].id

    @property
    def messages(self):
        """
        a default :class:`MessageIterator` for the channel
        """
        return self.messages_iter()

    def messages_iter(self, **kwargs):
        """
        Creates a new :class:`MessageIterator` for the channel with the given
        keyword arguments
        """
        return MessageIterator(self.client, self.id, before=self.last_message_id, **kwargs)

    @cached_property
    def guild(self):
        """
        Guild this channel belongs to (if relevant)
        """
        return self.client.state.guilds.get(self.guild_id)

    def get_invites(self):
        """
        Returns
        -------
        list(:class:`disco.types.invite.Invite`)
            All invites for this channel.
        """
        return self.client.api.channels_invites_list(self.id)

    def get_pins(self):
        """
        Returns
        -------
        list(:class:`disco.types.message.Message`)
            All pinned messages for this channel.
        """
        return self.client.api.channels_pins_list(self.id)

    def send_message(self, content, nonce=None, tts=False):
        """
        Send a message in this channel

        Parameters
        ----------
        content : str
            The message contents to send.
        nonce : Optional[snowflake]
            The nonce to attach to the message.
        tts : Optional[bool]
            Whether this is a TTS message.

        Returns
        -------
        :class:`disco.types.message.Message`
            The created message.
        """
        return self.client.api.channels_messages_create(self.id, content, nonce, tts)

    def connect(self, *args, **kwargs):
        """
        Connect to this channel over voice
        """
        assert self.is_voice, 'Channel must support voice to connect'
        vc = VoiceClient(self)
        vc.connect(*args, **kwargs)
        return vc


class MessageIterator(object):
    """
    An iterator which supports scanning through the messages for a channel.

    Parameters
    ----------
    client : :class:`disco.client.Client`
        The disco client instance to use when making requests.
    channel : `Channel`
        The channel to iterate within.
    direction : :attr:`MessageIterator.Direction`
        The direction in which this iterator will move.
    bulk : bool
        If true, this iterator will yield messages in list batches, otherwise each
        message will be yield individually.
    before : snowflake
        The message to begin scanning at.
    after : snowflake
        The message to begin scanning at.
    chunk_size : int
        The number of messages to request per API call.
    """
    Direction = Enum('UP', 'DOWN')

    def __init__(self, client, channel, direction=Direction.UP, bulk=False, before=None, after=None, chunk_size=100):
        self.client = client
        self.channel = channel
        self.direction = direction
        self.bulk = bulk
        self.before = before
        self.after = after
        self.chunk_size = chunk_size

        self.last = None
        self._buffer = []

        if not before and not after:
            raise Exception('Must specify at most one of before or after')

        if not any((before, after)) and self.direction == self.Direction.DOWN:
            raise Exception('Must specify either before or after for downward seeking')

    def fill(self):
        """
        Fills the internal buffer up with :class:`disco.types.message.Message` objects from the API
        """
        self._buffer = self.client.api.channels_messages_list(
                self.channel,
                before=self.before,
                after=self.after,
                limit=self.chunk_size)

        if not len(self._buffer):
            raise StopIteration

        self.after = None
        self.before = None

        if self.direction == self.Direction.UP:
            self.before = self._buffer[-1].id
        else:
            self._buffer.reverse()
            self.after == self._buffer[-1].id

    def next(self):
        return self.__next__()

    def __iter__(self):
        return self

    def __next__(self):
        if not len(self._buffer):
            self.fill()

        if self.bulk:
            res = self._buffer
            self._buffer = []
            return res
        else:
            return self._buffer.pop()



import re

from holster.enum import Enum

from disco.types.base import Model, snowflake, text, datetime, dictof, listof, enum
from disco.util import to_snowflake
from disco.util.functional import cached_property
from disco.types.user import User


MessageType = Enum(
    DEFAULT=0,
    RECIPIENT_ADD=1,
    RECIPIENT_REMOVE=2,
    CALL=3,
    CHANNEL_NAME_CHANGE=4,
    CHANNEL_ICON_CHANGE=5,
    PINS_ADD=6
)


class MessageEmbed(Model):
    """
    Message embed object

    Attributes
    ----------
    title : str
        Title of the embed.
    type : str
        Type of the embed.
    description : str
        Description of the embed.
    url : str
        URL of the embed.
    """
    title = text
    type = str
    description = text
    url = str


class MessageAttachment(Model):
    """
    Message attachment object

    Attributes
    ----------
    id : snowflake
        The id of this attachment.
    filename : str
        The filename of this attachment.
    url : str
        The URL of this attachment.
    proxy_url : str
        The URL to proxy through when downloading the attachment.
    size : int
        Size of the attachment.
    height : int
        Height of the attachment.
    width : int
        Width of the attachment.
    """
    id = str
    filename = text
    url = str
    proxy_url = str
    size = int
    height = int
    width = int


class Message(Model):
    """
    Represents a Message created within a Channel on Discord.

    Attributes
    ----------
    id : snowflake
        The ID of this message.
    channel_id : snowflake
        The channel ID this message was sent in.
    type : ``MessageType``
        Type of the message.
    author : :class:`disco.types.user.User`
        The author of this message.
    content : str
        The unicode contents of this message.
    nonce : str
        The nonce of this message.
    timestamp : datetime
        When this message was created.
    edited_timestamp : Optional[datetime]
        When this message was last edited.
    tts : bool
        Whether this is a TTS (text-to-speech) message.
    mention_everyone : bool
        Whether this message has an @everyone which mentions everyone.
    pinned : bool
        Whether this message is pinned in the channel.
    mentions : dict(snowflake, :class:`disco.types.user.User`)
        All users mentioned within this message.
    mention_roles : list(snowflake)
        All roles mentioned within this message.
    embeds : list(:class:`MessageEmbed`)
        All embeds for this message.
    attachments : list(:class:`MessageAttachment`)
        All attachments for this message.
    """
    id = snowflake
    channel_id = snowflake
    type = enum(MessageType)
    author = User
    content = text
    nonce = snowflake
    timestamp = datetime
    edited_timestamp = datetime
    tts = bool
    mention_everyone = bool
    pinned = bool
    mentions = dictof(User, key='id')
    mention_roles = listof(snowflake)
    embeds = listof(MessageEmbed)
    attachments = dictof(MessageAttachment, key='id')

    def __str__(self):
        return '<Message {} ({})>'.format(self.id, self.channel_id)

    @cached_property
    def guild(self):
        """
        Returns
        -------
        :class:`disco.types.guild.Guild`
            The guild (if applicable) this message was created in.
        """
        return self.channel.guild

    @cached_property
    def channel(self):
        """
        Returns
        -------
        :class:`disco.types.channel.Channel`
            The channel this message was created in.
        """
        return self.client.state.channels.get(self.channel_id)

    def reply(self, *args, **kwargs):
        """
        Reply to this message (proxys arguments to
        :func:`disco.types.channel.Channel.send_message`)

        Returns
        -------
        :class:`Message`
            The created message object.
        """
        return self.channel.send_message(*args, **kwargs)

    def edit(self, content):
        """
        Edit this message

        Args
        ----
        content : str
            The new edited contents of the message.

        Returns
        -------
        :class:`Message`
            The edited message object.
        """
        return self.client.api.channels_messages_modify(self.channel_id, self.id, content)

    def delete(self):
        """
        Delete this message.

        Returns
        -------
        :class:`Message`
            The deleted message object.
        """
        return self.client.api.channels_messages_delete(self.channel_id, self.id)

    def is_mentioned(self, entity):
        """
        Returns
        -------
        bool
            Whether the give entity was mentioned.
        """
        id = to_snowflake(entity)
        return id in self.mentions or id in self.mention_roles

    @cached_property
    def without_mentions(self):
        """
        Returns
        -------
        str
            the message contents with all valid mentions removed.
        """
        return self.replace_mentions(
            lambda u: '',
            lambda r: '')

    def replace_mentions(self, user_replace, role_replace):
        """
        Replaces user and role mentions with the result of a given lambda/function.

        Args
        ----
        user_replace : function
            A function taking a single argument, the user object mentioned, and
            returning a valid string.
        role_replace : function
            A function taking a single argument, the role ID mentioned, and
            returning a valid string.

        Returns
        -------
        str
            The message contents with all valid mentions replaced.
        """
        if not self.mentions and not self.mention_roles:
            return

        def replace(match):
            id = match.group(0)
            if id in self.mention_roles:
                return role_replace(id)
            else:
                return user_replace(self.mentions.get(id))

        return re.sub('<@!?([0-9]+)>', replace, self.content)



from __future__ import absolute_import

import logging


class LoggingClass(object):
    def __init__(self):
        self.log = logging.getLogger(self.__class__.__name__)

    def log_on_error(self, msg, f):
        def _f(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except:
                self.log.exception(msg)
                raise
        return _f



from __future__ import absolute_import, print_function

import six

try:
    import ujson as json
except ImportError:
    import json

from disco.gateway.encoding.base import BaseEncoder


class JSONEncoder(BaseEncoder):
    TYPE = 'json'

    @staticmethod
    def encode(obj):
        return json.dumps(obj)

    @staticmethod
    def decode(obj):
        return json.loads(obj)



from websocket import ABNF
from erlpack import unpack, pack

from disco.gateway.encoding.base import BaseEncoder


class ETFEncoder(BaseEncoder):
    TYPE = 'etf'
    OPCODE = ABNF.OPCODE_BINARY

    @staticmethod
    def encode(obj):
        return pack(obj)

    @staticmethod
    def decode(obj):
        return unpack(obj)



from holster.enum import Enum

from disco.api.http import APIException
from disco.util import to_snowflake
from disco.types.base import Model, snowflake, listof, dictof, datetime, text, binary, enum
from disco.types.user import User
from disco.types.voice import VoiceState
from disco.types.permissions import PermissionValue, Permissions, Permissible
from disco.types.channel import Channel


VerificationLevel = Enum(
    NONE=0,
    LOW=1,
    MEDIUM=2,
    HIGH=3,
    EXTREME=4,
)


class Emoji(Model):
    """
    An emoji object

    Attributes
    ----------
    id : snowflake
        The ID of this emoji.
    name : str
        The name of this emoji.
    require_colons : bool
        Whether this emoji requires colons to use.
    managed : bool
        Whether this emoji is managed by an integration.
    roles : list(snowflake)
        Roles this emoji is attached to.
    """
    id = snowflake
    name = text
    require_colons = bool
    managed = bool
    roles = listof(snowflake)


class Role(Model):
    """
    A role object

    Attributes
    ----------
    id : snowflake
        The role ID.
    name : string
        The role name.
    hoist : bool
        Whether this role is hoisted (displayed separately in the sidebar).
    managed : bool
        Whether this role is managed by an integration.
    color : int
        The RGB color of this role.
    permissions : :class:`disco.types.permissions.PermissionsValue`
        The permissions this role grants.
    position : int
        The position of this role in the hierarchy.
    """
    id = snowflake
    name = text
    hoist = bool
    managed = bool
    color = int
    permissions = PermissionValue
    position = int


class GuildMember(Model):
    """
    A GuildMember object

    Attributes
    ----------
    user : :class:`disco.types.user.User`
        The user object of this member.
    guild_id : snowflake
        The guild this member is part of.
    mute : bool
        Whether this member is server voice-muted.
    deaf : bool
        Whether this member is server voice-deafend.
    joined_at : datetime
        When this user joined the guild.
    roles : list(snowflake)
        Roles this member is part of.
    """
    user = User
    guild_id = snowflake
    mute = bool
    deaf = bool
    joined_at = datetime
    roles = listof(snowflake)

    def get_voice_state(self):
        """
        Returns
        -------
        Optional[:class:`disco.types.voice.VoiceState`]
            Returns the voice state for the member if they are currently connected
            to the guilds voice server.
        """
        return self.guild.get_voice_state(self)

    def kick(self):
        """
        Kicks the member from the guild.
        """
        self.client.api.guilds_members_kick(self.guild.id, self.user.id)

    def ban(self, delete_message_days=0):
        """
        Bans the member from the guild.

        Args
        ----
        delete_message_days : int
            The number of days to retroactively delete messages for.
        """
        self.client.api.guilds_bans_create(self.guild.id, self.user.id, delete_message_days)

    @property
    def id(self):
        """
        Alias to the guild members user id
        """
        return self.user.id


class Guild(Model, Permissible):
    """
    A guild object

    Attributes
    ----------
    id : snowflake
        The id of this guild.
    owner_id : snowflake
        The id of the owner.
    afk_channel_id : snowflake
        The id of the afk channel.
    embed_channel_id : snowflake
        The id of the embed channel.
    name : str
        Guilds name.
    icon : str
        Guilds icon (as PNG binary data).
    splash : str
        Guilds splash image (as PNG binary data).
    region : str
        Voice region.
    afk_timeout : int
        Delay after which users are automatically moved to the afk channel.
    embed_enabled : bool
        Whether the guilds embed is enabled.
    verification_level : int
        The verification level used by the guild.
    mfa_level : int
        The MFA level used by the guild.
    features : list(str)
        Extra features enabled for this guild.
    members : dict(snowflake, :class:`GuildMember`)
        All of the guilds members.
    channels : dict(snowflake, :class:`disco.types.channel.Channel`)
        All of the guilds channels.
    roles : dict(snowflake, :class:`Role`)
        All of the guilds roles.
    emojis : dict(snowflake, :class:`Emoji`)
        All of the guilds emojis.
    voice_states : dict(str, :class:`disco.types.voice.VoiceState`)
        All of the guilds voice states.
    """

    id = snowflake
    owner_id = snowflake
    afk_channel_id = snowflake
    embed_channel_id = snowflake
    name = text
    icon = binary
    splash = binary
    region = str
    afk_timeout = int
    embed_enabled = bool
    verification_level = enum(VerificationLevel)
    mfa_level = int
    features = listof(str)
    members = dictof(GuildMember, key='id')
    channels = dictof(Channel, key='id')
    roles = dictof(Role, key='id')
    emojis = dictof(Emoji, key='id')
    voice_states = dictof(VoiceState, key='session_id')

    def get_permissions(self, user):
        """
        Get the permissions a user has in this guild.

        Returns
        -------
        :class:`disco.types.permissions.PermissionValue`
            Computed permission value for the user.
        """
        if self.owner_id == user.id:
            return PermissionValue(Permissions.ADMINISTRATOR)

        member = self.get_member(user)
        value = PermissionValue(self.roles.get(self.id).permissions)

        for role in map(self.roles.get, member.roles):
            value += role.permissions

        return value

    def get_voice_state(self, user):
        """
        Attempt to get a voice state for a given user (who should be a member of
        this guild).

        Returns
        -------
        :class:`disco.types.voice.VoiceState`
            The voice state for the user in this guild.
        """
        user = to_snowflake(user)

        for state in self.voice_states.values():
            if state.user_id == user:
                return state

    def get_member(self, user):
        """
        Attempt to get a member from a given user.

        Returns
        -------
        :class:`GuildMember`
            The guild member object for the given user.
        """
        user = to_snowflake(user)

        if user not in self.members:
            try:
                self.members[user] = self.client.api.guilds_members_get(self.id, user)
            except APIException:
                return

        return self.members.get(user)

    def validate_members(self, ctx):
        if self.members:
            for member in self.members.values():
                member.guild = self
                member.guild_id = self.id

    def validate_channels(self, ctx):
        if self.channels:
            for channel in self.channels.values():
                channel.guild_id = self.id
                channel.guild = self



import gevent
import socket
import struct
import time

from holster.enum import Enum
from holster.emitter import Emitter

from disco.gateway.encoding.json import JSONEncoder
from disco.util.websocket import Websocket
from disco.util.logging import LoggingClass
from disco.voice.packets import VoiceOPCode
from disco.gateway.packets import OPCode

VoiceState = Enum(
    DISCONNECTED=0,
    AWAITING_ENDPOINT=1,
    AUTHENTICATING=2,
    CONNECTING=3,
    CONNECTED=4,
    VOICE_CONNECTING=5,
    VOICE_CONNECTED=6,
)

# TODO:
#   - player implementation
#   - encryption
#   - cleanup


class VoiceException(Exception):
    def __init__(self, msg, client):
        self.voice_client = client
        super(VoiceException, self).__init__(msg)


class UDPVoiceClient(LoggingClass):
    def __init__(self, vc):
        super(UDPVoiceClient, self).__init__()
        self.vc = vc
        self.conn = None
        self.ip = None
        self.port = None
        self.run_task = None
        self.connected = False

    def run(self):
        while True:
            self.conn.recvfrom(4096)

    def send(self, data):
        self.conn.sendto(data, (self.ip, self.port))

    def disconnect(self):
        self.run_task.kill()

    def connect(self, host, port, timeout=10):
        self.ip = socket.gethostbyname(host)
        self.port = port

        self.conn = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        # Send discovery packet
        packet = bytearray(70)
        struct.pack_into('>I', packet, 0, self.vc.ssrc)
        self.send(packet)

        # Wait for a response
        try:
            data, addr = gevent.spawn(lambda: self.conn.recvfrom(70)).get(timeout=timeout)
        except gevent.Timeout:
            return (None, None)

        # Read IP and port
        ip = str(data[4:]).split('\x00', 1)[0]
        port = struct.unpack('<H', data[-2:])[0]

        # Spawn read thread so we don't max buffers
        self.connected = True
        self.run_task = gevent.spawn(self.run)

        return (ip, port)


class VoiceClient(LoggingClass):
    def __init__(self, channel, encoder=None):
        super(VoiceClient, self).__init__()

        assert channel.is_voice, 'Cannot spawn a VoiceClient for a non-voice channel'
        self.channel = channel
        self.client = self.channel.client
        self.encoder = encoder or JSONEncoder

        self.packets = Emitter(gevent.spawn)
        self.packets.on(VoiceOPCode.READY, self.on_voice_ready)
        self.packets.on(VoiceOPCode.SESSION_DESCRIPTION, self.on_voice_sdp)

        # State
        self.state = VoiceState.DISCONNECTED
        self.connected = gevent.event.Event()
        self.token = None
        self.endpoint = None
        self.ssrc = None
        self.port = None

        self.update_listener = None

        # Websocket connection
        self.ws = None
        self.heartbeat_task = None

    def heartbeat(self, interval):
        while True:
            self.send(VoiceOPCode.HEARTBEAT, time.time() * 1000)
            gevent.sleep(interval / 1000)

    def set_speaking(self, value):
        self.send(VoiceOPCode.SPEAKING, {
            'speaking': value,
            'delay': 0,
        })

    def send(self, op, data):
        self.ws.send(self.encoder.encode({
            'op': op.value,
            'd': data,
        }), self.encoder.OPCODE)

    def on_voice_ready(self, data):
        self.state = VoiceState.CONNECTING
        self.ssrc = data['ssrc']
        self.port = data['port']

        self.heartbeat_task = gevent.spawn(self.heartbeat, data['heartbeat_interval'])

        self.udp = UDPVoiceClient(self)
        ip, port = self.udp.connect(self.endpoint, self.port)

        if not ip:
            self.disconnect()
            return

        self.send(VoiceOPCode.SELECT_PROTOCOL, {
            'protocol': 'udp',
            'data': {
                'port': port,
                'address': ip,
                'mode': 'plain'
            }
        })

    def on_voice_sdp(self, data):
        # Toggle speaking state so clients learn of our SSRC
        self.set_speaking(True)
        self.set_speaking(False)
        gevent.sleep(0.25)

        self.state = VoiceState.CONNECTED
        self.connected.set()

    def on_voice_server_update(self, data):
        if self.channel.guild_id != data.guild_id or not data.token:
            return

        if self.token and self.token != data.token:
            return

        self.token = data.token
        self.state = VoiceState.AUTHENTICATING

        self.endpoint = data.endpoint.split(':', 1)[0]
        self.ws = Websocket(
            'wss://' + self.endpoint,
            on_message=self.on_message,
            on_error=self.on_error,
            on_open=self.on_open,
            on_close=self.on_close,
        )
        self.ws.run_forever()

    def on_message(self, ws, msg):
        try:
            data = self.encoder.decode(msg)
        except:
            self.log.exception('Failed to parse voice gateway message: ')

        self.packets.emit(VoiceOPCode[data['op']], data['d'])

    def on_error(self, ws, err):
        # TODO
        self.log.warning('Voice websocket error: {}'.format(err))

    def on_open(self, ws):
        self.send(VoiceOPCode.IDENTIFY, {
            'server_id': self.channel.guild_id,
            'user_id': self.client.state.me.id,
            'session_id': self.client.gw.session_id,
            'token': self.token
        })

    def on_close(self, ws, code, error):
        # TODO
        self.log.warning('Voice websocket disconnected (%s, %s)', code, error)

    def connect(self, timeout=5, mute=False, deaf=False):
        self.state = VoiceState.AWAITING_ENDPOINT

        self.update_listener = self.client.events.on('VoiceServerUpdate', self.on_voice_server_update)

        self.client.gw.send(OPCode.VOICE_STATE_UPDATE, {
            'self_mute': mute,
            'self_deaf': deaf,
            'guild_id': int(self.channel.guild_id),
            'channel_id': int(self.channel.id),
        })

        if not self.connected.wait(timeout) or self.state != VoiceState.CONNECTED:
            raise VoiceException('Failed to connect to voice', self)

    def disconnect(self):
        self.state = VoiceState.DISCONNECTED

        if self.heartbeat_task:
            self.heartbeat_task.kill()
            self.heartbeat_task = None

        if self.ws and self.ws.sock.connected:
            self.ws.close()

        if self.udp and self.udp.connected:
            self.udp.disconnect()

        self.client.gw.send(OPCode.VOICE_STATE_UPDATE, {
            'self_mute': False,
            'self_deaf': False,
            'guild_id': int(self.channel.guild_id),
            'channel_id': None,
        })



import re

from disco.bot.command import CommandEvent


class BotConfig(object):
    """
    An object which is used to configure and define the runtime configuration for
    a bot.

    Attributes
    ----------
    token : str
        The authentication token for this bot. This is passed on to the
        :class:`disco.client.Client` without any validation.
    commands_enabled : bool
        Whether this bot instance should utilize command parsing. Generally this
        should be true, unless your bot is only handling events and has no user
        interaction.
    commands_require_mention : bool
        Whether messages must mention the bot to be considered for command parsing.
    commands_mention_rules : dict(str, bool)
        A dictionary describing what mention types can be considered a mention
        of the bot when using :attr:`commands_require_mention`. This dictionary
        can contain the following keys: `here`, `everyone`, `role`, `user`. When
        a keys value is set to true, the mention type will be considered for
        command parsing.
    commands_prefix : str
        A string prefix that is required for a message to be considered for
        command parsing.
    commands_allow_edit : bool
        If true, the bot will reparse an edited message if it was the last sent
        message in a channel, and did not previously trigger a command. This is
        helpful for allowing edits to typod commands.
    plugin_config_provider : Optional[function]
        If set, this function will be called before loading a plugin, with the
        plugins name. Its expected to return a type of configuration object the
        plugin understands.
    """
    token = None

    commands_enabled = True
    commands_require_mention = True
    commands_mention_rules = {
        # 'here': False,
        'everyone': False,
        'role': True,
        'user': True,
    }
    commands_prefix = ''
    commands_allow_edit = True

    plugin_config_provider = None


class Bot(object):
    """
    Disco's implementation of a simple but extendable Discord bot. Bots consist
    of a set of plugins, and a Disco client.

    Parameters
    ----------
    client : :class:`disco.client.Client`
        The client this bot should utilize for its connection.
    config : Optional[:class:`BotConfig`]
        The configuration to use for this bot. If not provided will use the defaults
        inside of :class:`BotConfig`.

    Attributes
    ----------
    client : `disco.client.Client`
        The client instance for this bot.
    config : `BotConfig`
        The bot configuration instance for this bot.
    plugins : dict(str, :class:`disco.bot.plugin.Plugin`)
        Any plugins this bot has loaded
    """
    def __init__(self, client, config=None):
        self.client = client
        self.config = config or BotConfig()

        if self.client.config.manhole_enable:
            self.client.manhole_locals['bot'] = self

        self.plugins = {}

        # Only bind event listeners if we're going to parse commands
        if self.config.commands_enabled:
            self.client.events.on('MessageCreate', self.on_message_create)

            if self.config.commands_allow_edit:
                self.client.events.on('MessageUpdate', self.on_message_update)

        # Stores the last message for every single channel
        self.last_message_cache = {}

        # Stores a giant regex matcher for all commands
        self.command_matches_re = None

    @classmethod
    def from_cli(cls, *plugins):
        """
        Creates a new instance of the bot using the utilities inside of the
        :mod:`disco.cli` module. Allows passing in a set of uninitialized
        plugin classes to load.

        Parameters
        ---------
        plugins : Optional[list(:class:`disco.bot.plugin.Plugin`)]
            Any plugins to load after creating the new bot instance

        """
        from disco.cli import disco_main
        inst = cls(disco_main())

        for plugin in plugins:
            inst.add_plugin(plugin)

        return inst

    @property
    def commands(self):
        """
        Generator of all commands this bots plugins have defined
        """
        for plugin in self.plugins.values():
            for command in plugin.commands.values():
                yield command

    def compute_command_matches_re(self):
        """
        Computes a single regex which matches all possible command combinations.
        """
        re_str = '|'.join(command.regex for command in self.commands)
        if re_str:
            self.command_matches_re = re.compile(re_str)
        else:
            self.command_matches_re = None

    def get_commands_for_message(self, msg):
        """
        Generator of all commands that a given message object triggers, based on
        the bots plugins and configuration.

        Parameters
        ---------
        msg : :class:`disco.types.message.Message`
            The message object to parse and find matching commands for

        Yields
        -------
        tuple(:class:`disco.bot.command.Command`, `re.MatchObject`)
            All commands the message triggers
        """
        content = msg.content

        if self.config.commands_require_mention:
            match = any((
                self.config.commands_mention_rules['user'] and msg.is_mentioned(self.client.state.me),
                self.config.commands_mention_rules['everyone'] and msg.mention_everyone,
                self.config.commands_mention_rules['role'] and any(map(msg.is_mentioned,
                    msg.guild.get_member(self.client.state.me).roles
                ))))

            if not match:
                raise StopIteration

            content = msg.without_mentions.strip()

        if self.config.commands_prefix and not content.startswith(self.config.commands_prefix):
            raise StopIteration
        else:
            content = content[len(self.config.commands_prefix):]

        if not self.command_matches_re or not self.command_matches_re.match(content):
            raise StopIteration

        for command in self.commands:
            match = command.compiled_regex.match(content)
            if match:
                yield (command, match)

    def handle_message(self, msg):
        """
        Attempts to handle a newly created or edited message in the context of
        command parsing/triggering. Calls all relevant commands the message triggers.

        Parameters
        ---------
        msg : :class:`disco.types.message.Message`
            The newly created or updated message object to parse/handle.

        Returns
        -------
        bool
            whether any commands where successfully triggered by the message
        """
        commands = list(self.get_commands_for_message(msg))

        if len(commands):
            return any([
                command.plugin.execute(CommandEvent(command, msg, match))
                for command, match in commands
            ])

        return False

    def on_message_create(self, event):
        if event.message.author.id == self.client.state.me.id:
            return

        if self.config.commands_allow_edit:
            self.last_message_cache[event.message.channel_id] = (event.message, False)

        self.handle_message(event.message)

    def on_message_update(self, event):
        if self.config.commands_allow_edit:
            obj = self.last_message_cache.get(event.message.channel_id)
            if not obj:
                return

            msg, triggered = obj
            if msg.id == event.message.id and not triggered:
                msg.update(event.message)
                triggered = self.handle_message(msg)

                self.last_message_cache[msg.channel_id] = (msg, triggered)

    def add_plugin(self, cls):
        """
        Adds and loads a plugin, based on its class.

        Parameters
        ----------
        cls : subclass of :class:`disco.bot.plugin.Plugin`
            Plugin class to initialize and load.
        """
        if cls.__name__ in self.plugins:
            raise Exception('Cannot add already added plugin: {}'.format(cls.__name__))

        config = self.config.plugin_config_provider(cls.__name__) if self.config.plugin_config_provider else None

        self.plugins[cls.__name__] = cls(self, config)
        self.plugins[cls.__name__].load()
        self.compute_command_matches_re()

    def rmv_plugin(self, cls):
        """
        Unloads and removes a plugin based on its class.

        Parameters
        ----------
        cls : subclass of :class:`disco.bot.plugin.Plugin`
            Plugin class to unload and remove.
        """
        if cls.__name__ not in self.plugins:
            raise Exception('Cannot remove non-existant plugin: {}'.format(cls.__name__))

        self.plugins[cls.__name__].unload()
        self.plugins[cls.__name__].destroy()
        del self.plugins[cls.__name__]
        self.compute_command_matches_re()

    def run_forever(self):
        """
        Runs this bots core loop forever
        """
        self.client.run_forever()



"""
The CLI module is a small utility that can be used as an easy entry point for
creating and running bots/clients.
"""
from __future__ import print_function

import logging
import argparse

from gevent import monkey

monkey.patch_all()

parser = argparse.ArgumentParser()
parser.add_argument('--token', help='Bot Authentication Token', required=True)
parser.add_argument('--shard-count', help='Total number of shards', default=1)
parser.add_argument('--shard-id', help='Current shard number/id', default=0)
parser.add_argument('--manhole', action='store_true', help='Enable the manhole', default=False)
parser.add_argument('--manhole-bind', help='host:port for the manhole to bind too', default='localhost:8484')
parser.add_argument('--encoder', help='encoder for gateway data', default='json')

logging.basicConfig(level=logging.INFO)


def disco_main():
    """
    Creates an argument parser and parses a standard set of command line arguments,
    creating a new :class:`Client`.

    Returns
    -------
    :class:`Client`
        A new Client from the provided command line arguments
    """
    args = parser.parse_args()

    from disco.client import Client, ClientConfig
    from disco.gateway.encoding import ENCODERS
    from disco.util.token import is_valid_token

    if not is_valid_token(args.token):
        print('Invalid token passed')
        return

    cfg = ClientConfig()
    cfg.token = args.token
    cfg.shard_id = args.shard_id
    cfg.shard_count = args.shard_count
    cfg.manhole_enable = args.manhole
    cfg.manhole_bind = args.manhole_bind
    cfg.encoding_cls = ENCODERS[args.encoder]

    return Client(cfg)

if __name__ == '__main__':
    disco_main().run_forever()



import re

TOKEN_RE = re.compile(r'M\w{23}\.[\w-]{6}\..{27}')


def is_valid_token(token):
    """
    Validates a Discord authentication token, returning true if valid
    """
    return bool(TOKEN_RE.match(token))



from holster.enum import Enum, EnumAttr

Permissions = Enum(
    CREATE_INSTANT_INVITE=1 << 0,
    KICK_MEMBERS=1 << 1,
    BAN_MEMBERS=1 << 2,
    ADMINISTRATOR=1 << 3,
    MANAGE_CHANNELS=1 << 4,
    MANAGE_GUILD=1 << 5,
    READ_MESSAGES=1 << 10,
    SEND_MESSAGES=1 << 11,
    SEND_TSS_MESSAGES=1 << 12,
    MANAGE_MESSAGES=1 << 13,
    EMBED_LINKS=1 << 14,
    ATTACH_FILES=1 << 15,
    READ_MESSAGE_HISTORY=1 << 16,
    MENTION_EVERYONE=1 << 17,
    USE_EXTERNAL_EMOJIS=1 << 18,
    CONNECT=1 << 20,
    SPEAK=1 << 21,
    MUTE_MEMBERS=1 << 22,
    DEAFEN_MEMBERS=1 << 23,
    MOVE_MEMBERS=1 << 24,
    USE_VAD=1 << 25,
    CHANGE_NICKNAME=1 << 26,
    MANAGE_NICKNAMES=1 << 27,
    MANAGE_ROLES=1 << 28,
    MANAGE_WEBHOOKS=1 << 29,
    MANAGE_EMOJIS=1 << 30,
)


class PermissionValue(object):
    def __init__(self, value=0):
        if isinstance(value, EnumAttr) or isinstance(value, PermissionValue):
            value = value.value

        self.value = value

    def can(self, *perms):
        for perm in perms:
            if isinstance(perm, EnumAttr):
                perm = perm.value
            if not (self.value & perm) == perm:
                return False
        return True

    def add(self, other):
        if isinstance(other, PermissionValue):
            self.value |= other.value
        elif isinstance(other, int):
            self.value |= other
        elif isinstance(other, EnumAttr):
            setattr(self, other.name, True)
        else:
            raise TypeError('Cannot PermissionValue.add from type {}'.format(type(other)))
        return self

    def sub(self, other):
        if isinstance(other, PermissionValue):
            self.value &= ~other.value
        elif isinstance(other, int):
            self.value &= other
        elif isinstance(other, EnumAttr):
            setattr(self, other.name, False)
        else:
            raise TypeError('Cannot PermissionValue.sub from type {}'.format(type(other)))
        return self

    def __iadd__(self, other):
        return self.add(other)

    def __isub__(self, other):
        return self.sub(other)

    def __getattribute__(self, name):
        if name in Permissions.attrs:
            return (self.value & Permissions[name].value) == Permissions[name].value
        else:
            return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        if name not in Permissions.attrs:
            return super(PermissionValue, self).__setattr__(name, value)

        if value:
            self.value |= Permissions[name].value
        else:
            self.value &= ~Permissions[name].value

    def to_dict(self):
        return {
            k: getattr(self, k) for k in Permissions.attrs
        }

    @classmethod
    def text(cls):
        return cls(523264)

    @classmethod
    def voice(cls):
        return cls(66060288)


class Permissible(object):
    def can(self, user, *args):
        perms = self.get_permissions(user)
        return perms.administrator or perms.can(*args)



from websocket import ABNF


class BaseEncoder(object):
    TYPE = None
    OPCODE = ABNF.OPCODE_TEXT

    @staticmethod
    def encode(obj):
        pass

    @staticmethod
    def decode(obj):
        pass



from gevent.backdoor import BackdoorServer


class DiscoBackdoorServer(BackdoorServer):
    def __init__(self, listener, localf=None, banner=None, **server_args):
        super(DiscoBackdoorServer, self).__init__(listener, {}, banner, **server_args)
        self.localf = localf

    def _create_interactive_locals(self):
        obj = super(DiscoBackdoorServer, self)._create_interactive_locals()
        obj.update(self.localf())
        return obj



from disco.types.base import Model, datetime
from disco.types.user import User
from disco.types.guild import Guild
from disco.types.channel import Channel


class Invite(Model):
    """
    An invite object

    Attributes
    ----------
    code : str
        The invite code.
    inviter : :class:`disco.types.user.User`
        The user who created this invite.
    guild : :class:`disco.types.guild.Guild`
        The guild this invite is for.
    channel : :class:`disco.types.channel.Channel`
        The channel this invite is for.
    max_age : int
        The time after this invites creation at which it expires.
    max_uses : int
        The maximum number of uses.
    uses : int
        The current number of times the invite was used.
    temporary : bool
        Whether this invite only grants temporary memborship.
    created_at : datetime
        When this invite was created.
    """
    code = str
    inviter = User
    guild = Guild
    channel = Channel
    max_age = int
    max_uses = int
    uses = int
    temporary = bool
    created_at = datetime



import gevent

from holster.emitter import Emitter

from disco.state import State
from disco.api.client import APIClient
from disco.gateway.client import GatewayClient
from disco.util.logging import LoggingClass
from disco.util.backdoor import DiscoBackdoorServer


class ClientConfig(LoggingClass):
    """
    Configuration for the :class:`Client`.

    Attributes
    ----------
    token : str
        Discord authentication token, ca be validated using the
        :func:`disco.util.token.is_valid_token` function.
    shard_id : int
        The shard ID for the current client instance.
    shard_count : int
        The total count of shards running.
    manhole_enable : bool
        Whether to enable the manhole (e.g. console backdoor server) utility.
    manhole_bind : tuple(str, int)
        A (host, port) combination which the manhole server will bind to (if its
        enabled using :attr:`manhole_enable`).
    encoding_cls : class
        The class to use for encoding/decoding data from websockets.
    """

    token = ""
    shard_id = 0
    shard_count = 1

    manhole_enable = True
    manhole_bind = ('127.0.0.1', 8484)

    encoding_cls = None


class Client(object):
    """
    Class representing the base entry point that should be used in almost all
    implementation cases. This class wraps the functionality of both the REST API
    (:class:`disco.api.client.APIClient`) and the realtime gateway API
    (:class:`disco.gateway.client.GatewayClient`).

    Parameters
    ----------
    config : :class:`ClientConfig`
        Configuration for this client instance.

    Attributes
    ----------
    config : :class:`ClientConfig`
        The runtime configuration for this client.
    events : :class:`Emitter`
        An emitter which emits Gateway events.
    packets : :class:`Emitter`
        An emitter which emits Gateway packets.
    state : :class:`State`
        The state tracking object.
    api : :class:`APIClient`
        The API client.
    gw : :class:`GatewayClient`
        The gateway client.
    manhole_locals : dict
        Dictionary of local variables for each manhole connection. This can be
        modified to add/modify local variables.
    manhole : Optional[:class:`BackdoorServer`]
        Gevent backdoor server (if the manhole is enabled).
    """
    def __init__(self, config):
        super(Client, self).__init__()
        self.config = config

        self.events = Emitter(gevent.spawn)
        self.packets = Emitter(gevent.spawn)

        self.state = State(self)
        self.api = APIClient(self)
        self.gw = GatewayClient(self, self.config.encoding_cls)

        if self.config.manhole_enable:
            self.manhole_locals = {}
            self.manhole = DiscoBackdoorServer(self.config.manhole_bind,
                banner='Disco Manhole',
                localf=lambda: self.manhole_locals)
            self.manhole.start()

    def run(self):
        """
        Run the client (e.g. the :class:`GatewayClient`) in a new greenlet
        """
        return gevent.spawn(self.gw.run)

    def run_forever(self):
        """
        Run the client (e.g. the :class:`GatewayClient`) in the current greenlet
        """
        return self.gw.run()



from setuptools import setup, find_packages

from disco import VERSION

with open('requirements.txt') as f:
    requirements = f.readlines()

with open('README.md') as f:
    readme = f.read()

setup(
    name='disco-py',
    author='b1nzy',
    url='https://github.com/b1naryth1ef/disco',
    version=VERSION,
    packages=find_packages(),
    license='MIT',
    description='A Python library for Discord',
    long_description=readme,
    include_package_data=True,
    install_requires=requirements,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ])



from disco.types.base import Model, snowflake, text, binary


class User(Model):
    id = snowflake
    username = text
    discriminator = str
    avatar = binary
    verified = bool
    email = str

    def to_string(self):
        return '{}#{}'.format(self.username, self.discriminator)

    def __str__(self):
        return '<User {} ({})>'.format(self.id, self.to_string())

    def on_create(self):
        self.client.state.users[self.id] = self



import six


def to_snowflake(i):
    if isinstance(i, six.integer_types):
        return i
    elif isinstance(i, str):
        return int(i)
    elif hasattr(i, 'id'):
        return i.id

    raise Exception('{} ({}) is not convertable to a snowflake'.format(type(i), i))



import inflection

from collections import defaultdict, deque, namedtuple
from weakref import WeakValueDictionary

from disco.gateway.packets import OPCode


class StackMessage(namedtuple('StackMessage', ['id', 'channel_id', 'author_id'])):
    """
    A message stored on a stack inside of the state object, used for tracking
    previously sent messages in channels.

    Attributes
    ---------
    id : snowflake
        the id of the message
    channel_id : snowflake
        the id of the channel this message was sent in
    author_id : snowflake
        the id of the author of this message
    """


class StateConfig(object):
    """
    A configuration object for determining how the State tracking behaves.

    Attributes
    ----------
    track_messages : bool
        Whether the state store should keep a buffer of previously sent messages.
        Message tracking allows for multiple higher-level shortcuts and can be
        highly useful when developing bots that need to delete their own messages.

        Message tracking is implemented using a deque and a namedtuple, meaning
        it should generally not have a high impact on memory, however users who
        find they do not need and may be experiencing memory pressure can disable
        this feature entirely using this attribute.
    track_messages_size : int
        The size of the deque for each channel. Using this you can calculate the
        total number of possible :class:`StackMessage` objects kept in memory,
        using: `total_mesages_size * total_channels`. This can be tweaked based
        on usage to help prevent memory pressure.
    """
    track_messages = True
    track_messages_size = 100


class State(object):
    """
    The State class is used to track global state based on events emitted from
    the :class:`GatewayClient`. State tracking is a core component of the Disco
    client, providing the mechanism for most of the higher-level utility functions.

    Attributes
    ----------
    EVENTS : list(str)
        A list of all events the State object binds too.
    client : :class:`disco.client.Client`
        The Client instance this state is attached too
    config : :class:`StateConfig`
        The configuration for this state instance
    me : :class:`disco.types.user.User`
        The currently logged in user
    dms : dict(snowflake, :class:`disco.types.channel.Channel`)
        Mapping of all known DM Channels
    guilds : dict(snowflake, :class:`disco.types.guild.Guild`)
        Mapping of all known/loaded Guilds
    channels : dict(snowflake, :class:`disco.types.channel.Channel`)
        Weak mapping of all known/loaded Channels
    users : dict(snowflake, :class:`disco.types.user.User`)
        Weak mapping of all known/loaded Users
    voice_states : dict(str, :class:`disco.types.voice.VoiceState`)
        Weak mapping of all known/active Voice States
    messages : Optional[dict(snowflake, :class:`deque`)]
        Mapping of channel ids to deques containing :class:`StackMessage` objects
    """
    EVENTS = [
        'Ready', 'GuildCreate', 'GuildUpdate', 'GuildDelete', 'GuildMemberAdd', 'GuildMemberRemove',
        'GuildMemberUpdate', 'GuildMembersChunk', 'GuildRoleCreate', 'GuildRoleUpdate', 'GuildRoleDelete',
        'ChannelCreate', 'ChannelUpdate', 'ChannelDelete', 'VoiceStateUpdate'
    ]

    def __init__(self, client, config=None):
        self.client = client
        self.config = config or StateConfig()

        self.me = None
        self.dms = {}
        self.guilds = {}
        self.channels = WeakValueDictionary()
        self.users = WeakValueDictionary()
        self.voice_states = WeakValueDictionary()

        # If message tracking is enabled, listen to those events
        if self.config.track_messages:
            self.messages = defaultdict(lambda: deque(maxlen=self.config.track_messages_size))
            self.EVENTS += ['MessageCreate', 'MessageDelete']

        # The bound listener objects
        self.listeners = []
        self.bind()

    def unbind(self):
        """
        Unbinds all bound event listeners for this state object
        """
        map(lambda k: k.unbind(), self.listeners)
        self.listeners = []

    def bind(self):
        """
        Binds all events for this state object, storing the listeners for later
        unbinding.
        """
        assert not len(self.listeners), 'Binding while already bound is dangerous'

        for event in self.EVENTS:
            func = 'on_' + inflection.underscore(event)
            self.listeners.append(self.client.events.on(event, getattr(self, func)))

    def on_ready(self, event):
        self.me = event.user

    def on_message_create(self, event):
        self.messages[event.message.channel_id].append(
            StackMessage(event.message.id, event.message.channel_id, event.message.author.id))

    def on_message_update(self, event):
        message, cid = event.message, event.message.channel_id
        if cid not in self.messages:
            return

        sm = next((i for i in self.messages[cid] if i.id == message.id), None)
        if not sm:
            return

        sm.id = message.id
        sm.channel_id = cid
        sm.author_id = message.author.id

    def on_message_delete(self, event):
        if event.channel_id not in self.messages:
            return

        sm = next((i for i in self.messages[event.channel_id] if i.id == event.id), None)
        if not sm:
            return

        self.messages[event.channel_id].remove(sm)

    def on_guild_create(self, event):
        self.guilds[event.guild.id] = event.guild
        self.channels.update(event.guild.channels)

        for channel in event.guild.channels.values():
            channel.guild_id = event.guild.id
            channel.guild = event.guild

        for member in event.guild.members.values():
            self.users[member.user.id] = member.user

        # Request full member list
        self.client.gw.send(OPCode.REQUEST_GUILD_MEMBERS, {
            'guild_id': event.guild.id,
            'query': '',
            'limit': 0,
        })

    def on_guild_update(self, event):
        self.guilds[event.guild.id].update(event.guild)

    def on_guild_delete(self, event):
        if event.guild_id in self.guilds:
            # Just delete the guild, channel references will fall
            del self.guilds[event.guild_id]

    def on_channel_create(self, event):
        if event.channel.is_guild and event.channel.guild_id in self.guilds:
            self.guilds[event.channel.guild_id].channels[event.channel.id] = event.channel
            self.channels[event.channel.id] = event.channel
        elif event.channel.is_dm:
            self.dms[event.channel.id] = event.channel
            self.channels[event.channel.id] = event.channel

    def on_channel_update(self, event):
        if event.channel.id in self.channels:
            self.channels[event.channel.id].update(event.channel)

    def on_channel_delete(self, event):
        if event.channel.is_guild and event.channel.guild_id in self.guilds:
            del self.guilds[event.channel.id]
        elif event.channel.is_dm:
            del self.pms[event.channel.id]

    def on_voice_state_update(self, event):
        # Happy path: we have the voice state and want to update/delete it
        guild = self.guilds.get(event.state.guild_id)

        if event.state.session_id in guild.voice_states:
            if event.state.channel_id:
                guild.voice_states[event.state.session_id].update(event.state)
            else:
                del guild.voice_states[event.state.session_id]
        elif event.state.channel_id:
            guild.voice_states[event.state.session_id] = event.state

    def on_guild_member_add(self, event):
        if event.member.user.id not in self.users:
            self.users[event.member.user.id] = event.member.user
        else:
            event.member.user = self.users[event.member.user.id]

        if event.member.guild_id not in self.guilds:
            return

        event.member.guild = self.guilds[event.member.guild_id]
        self.guilds[event.member.guild_id].members[event.member.id] = event.member

    def on_guild_member_update(self, event):
        if event.guild_id not in self.guilds:
            return

        self.guilds[event.guild_id].members[event.user.id].roles = event.roles
        self.guilds[event.guild_id].members[event.user.id].user.update(event.user)

    def on_guild_member_remove(self, event):
        if event.guild_id not in self.guilds:
            return

        if event.user.id not in self.guilds[event.guild_id].members:
            return

        del self.guilds[event.guild_id].members[event.user.id]

    def on_guild_members_chunk(self, event):
        if event.guild_id not in self.guilds:
            return

        guild = self.guilds[event.guild_id]
        for member in event.members:
            member.guild = guild
            member.guild_id = guild.id
            guild.members[member.id] = member
            self.users[member.id] = member.user

    def on_guild_role_create(self, event):
        if event.guild_id not in self.guilds:
            return

        self.guilds[event.guild_id].roles[event.role.id] = event.role

    def on_guild_role_update(self, event):
        if event.guild_id not in self.guilds:
            return

        self.guilds[event.guild_id].roles[event.role.id].update(event.role)

    def on_guild_role_delete(self, event):
        if event.guild_id not in self.guilds:
            return

        del self.guilds[event.guild_id].roles[event.role.id]



from holster.enum import Enum

OPCode = Enum(
    DISPATCH=0,
    HEARTBEAT=1,
    IDENTIFY=2,
    STATUS_UPDATE=3,
    VOICE_STATE_UPDATE=4,
    VOICE_SERVER_PING=5,
    RESUME=6,
    RECONNECT=7,
    REQUEST_GUILD_MEMBERS=8,
    INVALID_SESSION=9,
    HELLO=10,
    HEARTBEAT_ACK=11,
    GUILD_SYNC=12,
)



import re
import copy


# Regex which splits out argument parts
PARTS_RE = re.compile('(\<|\[)((?:\w+|\:|\||\.\.\.| (?:[0-9]+))+)(?:\>|\])')

# Mapping of types
TYPE_MAP = {
    'str': str,
    'int': int,
    'float': float,
    'snowflake': int,
}


class ArgumentError(Exception):
    """
    An error thrown when passed in arguments cannot be conformed/casted to the
    argument specification.
    """


class Argument(object):
    """
    A single argument, which is normally the member of a :class:`ArgumentSet`.

    Attributes
    ----------
    name : str
        The name of this argument.
    count : int
        The number of raw arguments that compose this argument.
    required : bool
        Whether this is a required argument.
    types : list(type)
        Types this argument supports.
    """
    def __init__(self, raw):
        self.name = None
        self.count = 1
        self.required = False
        self.types = None
        self.parse(raw)

    @property
    def true_count(self):
        """
        The true number of raw arguments this argument takes
        """
        return self.count or 1

    def parse(self, raw):
        """
        Attempts to parse arguments from their raw form
        """
        prefix, part = raw

        if prefix == '<':
            self.required = True
        else:
            self.required = False

        if part.endswith('...'):
            part = part[:-3]
            self.count = 0
        elif ' ' in part:
            split = part.split(' ', 1)
            part, self.count = split[0], int(split[1])

        if ':' in part:
            part, typeinfo = part.split(':')
            self.types = typeinfo.split('|')

        self.name = part.strip()


class ArgumentSet(object):
    """
    A set of :class:`Argument` instances which forms a larger argument specification

    Attributes
    ----------
    args : list(:class:`Argument`)
        All arguments that are a member of this set.
    types : dict(str, type)
        All types supported by this ArgumentSet.
    """
    def __init__(self, args=None, custom_types=None):
        self.args = args or []
        self.types = copy.copy(TYPE_MAP)
        self.types.update(custom_types or {})

    @classmethod
    def from_string(cls, line, custom_types=None):
        """
        Creates a new :class:`ArgumentSet` from a given argument string specification
        """
        args = cls(custom_types=custom_types)

        data = PARTS_RE.findall(line)
        if len(data):
            for item in data:
                args.append(Argument(item))

        return args

    def convert(self, types, value):
        """
        Attempts to convert a value to one or more types.

        Parameters
        ----------
        types : list(type)
            List of types to attempt conversion with.
        value : str
            The string value to attempt conversion on.
        """
        for typ_name in types:
            typ = self.types.get(typ_name)
            if not typ:
                raise Exception('Unknown type {}'.format(typ_name))

            try:
                return typ(value)
            except Exception as e:
                continue

        raise e

    def append(self, arg):
        """
        Add a new :class:`Argument` to this argument specification/set
        """
        if self.args and not self.args[-1].required and arg.required:
            raise Exception('Required argument cannot come after an optional argument')

        if self.args and not self.args[-1].count:
            raise Exception('No arguments can come after a catch-all')

        self.args.append(arg)

    def parse(self, rawargs):
        """
        Parse a string of raw arguments into this argument specification.
        """
        parsed = []

        for index, arg in enumerate(self.args):
            if not arg.required and index + arg.true_count > len(rawargs):
                continue

            if arg.count == 0:
                raw = rawargs[index:]
            else:
                raw = rawargs[index:index + arg.true_count]

            if arg.types:
                for idx, r in enumerate(raw):
                    try:
                        raw[idx] = self.convert(arg.types, r)
                    except:
                        raise ArgumentError('cannot convert `{}` to `{}`'.format(
                            r, ', '.join(arg.types)
                        ))

            if arg.count == 1:
                raw = raw[0]

            if (not arg.types or arg.types == ['str']) and isinstance(raw, list):
                raw = ' '.join(raw)

            parsed.append(raw)

        return parsed

    @property
    def length(self):
        """
        The number of arguments in this set/specification
        """
        return len(self.args)

    @property
    def required_length(self):
        """
        The number of required arguments to compile this set/specificaiton
        """
        return sum([i.true_count for i in self.args if i.required])



import six
import inspect
import functools

from datetime import datetime as real_datetime

DATETIME_FORMATS = [
    '%Y-%m-%dT%H:%M:%S.%f',
    '%Y-%m-%dT%H:%M:%S'
]


def _make(typ, data, client):
    args, _, _, _ = inspect.getargspec(typ)
    if 'client' in args:
        return typ(data, client)
    return typ(data)


def snowflake(data):
    return int(data) if data else None


def enum(typ):
    def _f(data):
        return typ.get(data) if data else None
    return _f


def listof(typ):
    def _f(data, client=None):
        if not data:
            return []
        return [_make(typ, obj, client) for obj in data]
    return _f


def dictof(typ, key=None):
    def _f(data, client=None):
        if not data:
            return {}

        if key:
            return {
                getattr(v, key): v for v in (
                    _make(typ, i, client) for i in data
                )}
        else:
            return {k: _make(typ, v, client) for k, v in six.iteritems(data)}
    return _f


def alias(typ, name):
    return ('alias', name, typ)


def datetime(data):
    if not data:
        return None

    for fmt in DATETIME_FORMATS:
        try:
            return real_datetime.strptime(data.rsplit('+', 1)[0], fmt)
        except (ValueError, TypeError):
            continue

    raise ValueError('Failed to conver `{}` to datetime'.format(data))


def text(obj):
    return six.text_type(obj) if obj else six.text_type()


def binary(obj):
    return six.text_type(obj) if obj else six.text_type()


class ModelMeta(type):
    def __new__(cls, name, parents, dct):
        fields = {}
        for k, v in six.iteritems(dct):
            if isinstance(v, tuple):
                if v[0] == 'alias':
                    fields[v[1]] = (k, v[2])
                    continue

            if inspect.isclass(v):
                fields[k] = v
            elif callable(v):
                args, _, _, _ = inspect.getargspec(v)
                if 'self' in args:
                    continue

                fields[k] = v

        dct['_fields'] = fields
        return super(ModelMeta, cls).__new__(cls, name, parents, dct)


class Model(six.with_metaclass(ModelMeta)):
    def __init__(self, obj, client=None):
        self.client = client

        for name, typ in self.__class__._fields.items():
            dest_name = name

            if isinstance(typ, tuple):
                dest_name, typ = typ

            if name not in obj or not obj[name]:
                if inspect.isclass(typ) and issubclass(typ, Model):
                    res = None
                elif isinstance(typ, type):
                    res = typ()
                else:
                    res = typ(None)
                setattr(self, dest_name, res)
                continue

            try:
                if client:
                    args, _, _, _ = inspect.getargspec(typ)
                    if 'client' in args:
                        v = typ(obj[name], client)
                    else:
                        v = typ(obj[name])
                else:
                    v = typ(obj[name])
            except Exception:
                print('Failed during parsing of field {} => {} (`{}`)'.format(name, typ, obj[name]))
                raise

            if client and isinstance(v, Model):
                v.client = client

            setattr(self, dest_name, v)

    def update(self, other):
        for name in six.iterkeys(self.__class__._fields):
            value = getattr(other, name)
            if value:
                setattr(self, name, value)

        # Clear cached properties
        for name in dir(type(self)):
            if isinstance(getattr(type(self), name), property):
                delattr(self, name)


    @classmethod
    def create(cls, client, data):
        return cls(data, client)

    @classmethod
    def create_map(cls, client, data):
        return list(map(functools.partial(cls.create, client), data))



from gevent.lock import RLock


def cached_property(f):
    """
    Creates a cached class property out of ``f``. When the property is resolved
    for the first time, the function will be called and its result will be cached.
    Subsequent calls will return the cached value. If this property is set, the
    cached value will be replaced (or set initially) with the value provided. If
    this property is deleted, the cache will be cleared and the next call will
    refill it with a new value.

    Notes
    -----
    This function is greenlet safe.

    Args
    ----
    f : function
        The function to wrap.

    Returns
    -------
    property
        The cached property created.
    """
    lock = RLock()
    value_name = '_' + f.__name__

    def getf(self, *args, **kwargs):
        if not hasattr(self, value_name):
            with lock:
                if hasattr(self, value_name):
                    return getattr(self, value_name)

                setattr(self, value_name, f(self, *args, **kwargs))

        return getattr(self, value_name)

    def setf(self, value):
        setattr(self, value_name, value)

    def delf(self):
        delattr(self, value_name)

    return property(getf, setf, delf)



from __future__ import absolute_import

import websocket
import gevent
import six

from holster.emitter import Emitter

from disco.util.logging import LoggingClass


class Websocket(LoggingClass, websocket.WebSocketApp):
    """
    A utility class which wraps the functionality of :class:`websocket.WebSocketApp`
    changing its behavior to better conform with standard style across disco.

    The major difference comes with the move from callback functions, to all
    events being piped into a single emitter.
    """
    def __init__(self, *args, **kwargs):
        LoggingClass.__init__(self)
        websocket.WebSocketApp.__init__(self, *args, **kwargs)

        self.emitter = Emitter(gevent.spawn)

        # Hack to get events to emit
        for var in self.__dict__.keys():
            if not var.startswith('on_'):
                continue

            setattr(self, var, var)

    def _get_close_args(self, data):
        if data and len(data) >= 2:
            code = 256 * six.byte2int(data[0:1]) + six.byte2int(data[1:2])
            reason = data[2:].decode('utf-8')
            return [code, reason]
        return [None, None]

    def _callback(self, callback, *args):
        if not callback:
            return

        self.emitter.emit(callback, *args)



import time
import gevent


class RouteState(object):
    """
    An object which stores ratelimit state for a given method/url route
    combination (as specified in :class:`disco.api.http.Routes`).

    Parameters
    ----------
    route : tuple(HTTPMethod, str)
        The route which this RouteState is for.
    response : :class:`requests.Response`
        The response object for the last request made to the route, should contain
        the standard rate limit headers.

    Attributes
    ---------
    route : tuple(HTTPMethod, str)
        The route which this RouteState is for.
    remaining : int
        The number of remaining requests to the route before the rate limit will
        be hit, triggering a 429 response.
    reset_time : int
        A unix epoch timestamp (in seconds) after which this rate limit is reset
    event : :class:`gevent.event.Event`
        An event that is used to block all requests while a route is in the
        cooldown stage.
    """
    def __init__(self, route, response):
        self.route = route
        self.remaining = 0
        self.reset_time = 0
        self.event = None

        self.update(response)

    @property
    def chilled(self):
        """
        Whether this route is currently being cooldown (aka waiting until reset_time)
        """
        return self.event is not None

    @property
    def next_will_ratelimit(self):
        """
        Whether the next request to the route (at this moment in time) will
        trigger the rate limit.
        """

        if self.remaining - 1 < 0 and time.time() <= self.reset_time:
            return True

        return False

    def update(self, response):
        """
        Updates this route with a given Requests response object. Its expected
        the response has the required headers, however in the case it doesn't
        this function has no effect.
        """
        if 'X-RateLimit-Remaining' not in response.headers:
            return

        self.remaining = int(response.headers.get('X-RateLimit-Remaining'))
        self.reset_time = int(response.headers.get('X-RateLimit-Reset'))

    def wait(self, timeout=None):
        """
        Waits until this route is no longer under a cooldown

        Parameters
        ----------
        timeout : Optional[int]
            A timeout (in seconds) after which we will give up waiting


        Returns
        -------
        bool
            False if the timeout period expired before the cooldown was finished
        """
        return self.event.wait(timeout)

    def cooldown(self):
        """
        Waits for the current route to be cooled-down (aka waiting until reset time)
        """
        if self.reset_time - time.time() < 0:
            raise Exception('Cannot cooldown for negative time period; check clock sync')

        self.event = gevent.event.Event()
        gevent.sleep((self.reset_time - time.time()) + .5)
        self.event.set()
        self.event = None


class RateLimiter(object):
    """
    A in-memory store of ratelimit states for all routes we've ever called.

    Attributes
    ----------
    states : dict(tuple(HTTPMethod, str), :class:`RouteState`)
        Contains a :class:`RouteState` for each route the RateLimiter is currently
        tracking.
    """
    def __init__(self):
        self.states = {}

    def check(self, route, timeout=None):
        """
        Checks whether a given route can be called. This function will return
        immediately if no rate-limit cooldown is being imposed for the given
        route, or will wait indefinently (unless timeout is specified) until
        the route is finished being cooled down. This function should be called
        before making a request to the specified route.

        Parameters
        ----------
        route : tuple(HTTPMethod, str)
            The route that will be checked.
        timeout : Optional[int]
            A timeout after which we'll give up waiting for a routes cooldown
            to expire, and immedietly return.

        Returns
        -------
        bool
            False if the timeout period expired before the route finished cooling
            down.
        """
        return self._check(None, timeout) and self._check(route, timeout)

    def _check(self, route, timeout=None):
        if route in self.states:
            # If we're current waiting, join the club
            if self.states[route].chilled:
                return self.states[route].wait(timeout)

            if self.states[route].next_will_ratelimit:
                gevent.spawn(self.states[route].cooldown).get(True, timeout)

        return True

    def update(self, route, response):
        """
        Updates the given routes state with the rate-limit headers inside the
        response from a previous call to the route.

        Parameters
        ---------
        route : tuple(HTTPMethod, str)
            The route that will be updated.
        response : :class:`requests.Response`
            The response object for the last request to the route, whose headers
            will be used to update the routes rate limit state.
        """
        if 'X-RateLimit-Global' in response.headers:
            route = None

        if route in self.states:
            self.states[route].update(response)
        else:
            self.states[route] = RouteState(route, response)
